package example.MDBean;

import java.rmi.RemoteException;
import java.util.Properties;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;



import javax.ejb.CreateException;
import javax.ejb.RemoveException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;

/**
 * This class illustrates calling a Message-Driven bean and pipes messages
 * onto a JMS topic.
 *
 * @author Surlu M. Rao
 */

public class Client {
  static private String TOPIC_NAME = "MsgPipeTopic";

  private String m_url;
  
  private Context m_context;
  private TopicConnection m_topicConnection;

  public Client(String url)
    throws NamingException
  {
    m_url = url;
        
    try {
      // Get an initial context.
      m_context = getInitialContext();
      
      // Create the connection and start it
      TopicConnectionFactory cf =
        (TopicConnectionFactory) m_context.lookup("javax.jms.TopicConnectionFactory");
      m_topicConnection = cf.createTopicConnection();
      m_topicConnection.start();
      
    }
    catch(Exception ex) {
      ex.printStackTrace();
    }
  }

  /**
   * Runs this example from the command line. Example:
   * <p>
   * <tt>java examples.ejb20.message.Client "t3://localhost:7001"</tt>
   * <p>
   * The parameters are mandatory, just to make sure it is generic.
   * <p>
   * @param url               URL such as "t3://localhost:7001" of Server
   * @param inputMsg          Message to be piped, such as "this is a piped Message"
   *                          Note : if the message is more than one word put it in double quotes.
   */
  public static void main(String[] args) throws Exception {

    log("Beginning main");

    String url       = null;
    String inputMsg  = null;
    
    // Parse the argument list 
     if (args.length != 2) {
      System.out.println("Usage: java example.MDBean.Client t3://hostname:port MessageToPipe");   
      return;
    } else {
      url = args[0];
      inputMsg = args[1];
    }

    Client client = null;
    try {
      client = new Client(url);
    } catch (NamingException ne) {
      System.exit(1);
    }

    try {
      client.pipeMessage(inputMsg);
    }
    catch (Exception e) {
      log("There was an exception while piping the message.");
      log("there was a problem communicating with the server: "+e);
    } 

    log("Ending main");
  }

  /**
   * The piping method.
   * 
   * @param msg     message to pipe
   */
  public void pipeMessage(String msg) throws RemoteException, JMSException, NamingException  {
    
    log(" Beginning pipeMessage method ");
    
    Topic newTopic = null;
    TopicSession session = null;
    try {
      session =
        m_topicConnection.createTopicSession(false,   // non transacted
                                             Session.AUTO_ACKNOWLEDGE);
      log(" Looking up topic in JNDI "+ TOPIC_NAME);
      newTopic = (Topic) m_context.lookup(TOPIC_NAME);
    }
    catch(NamingException ex) {
      log(" Creating topic because it is not found ");
      newTopic = session.createTopic(TOPIC_NAME);
      m_context.bind(TOPIC_NAME, newTopic);
      log(" Created and bound topic to JNDI successfully ");
    }

    log(" Creating TopicPublisher " );
    TopicPublisher sender = session.createPublisher(newTopic);
    TextMessage tm = session.createTextMessage();
    tm.setText(msg);
    sender.publish(tm);
    log(" Ending pipeMessage ");
    
  }


  /**
   * Using a Properties object will work on JDK 1.1.x and Java2
   * clients
   */
  private Context getInitialContext() throws NamingException {
    
    try {
      // Get an InitialContext
      Properties h = new Properties();
      h.put(Context.INITIAL_CONTEXT_FACTORY,
        "weblogic.jndi.WLInitialContextFactory");
      h.put(Context.PROVIDER_URL, m_url);
      return new InitialContext(h);
    }
    catch (NamingException ex) {
      log("Error connecting to the server at "+m_url);
      log("Please make sure that the server is running.");
      throw ex;
    }
  }

 /**
  * This method logs to sysout. 
  * Useful for tracing the program sequence apart from the errors.
  *
  * @param s    text to log
  */
  private static void log(String s) {
    System.out.println("Log : " + s);
  }
  
}
