/* bobheap.c - bob heap management routines */
/*
        Copyright (c) 2001, by David Michael Betz
        All rights reserved
*/

#include <stdlib.h>
#include <string.h>
#include "bob.h"

/* VALUE */

#define ValueSize(o)                    (BobQuickGetDispatch(o)->size(o))
#define ScanValue(c,o)                  (BobQuickGetDispatch(o)->scan(c,o))

/* prototypes */
static void InitInterpreter(BobInterpreter *c);
static BobMemorySpace *NewMemorySpace(BobInterpreter *c,size_t size);

/* BobMakeInterpreter - make a new interpreter */
BobInterpreter *BobMakeInterpreter(long size,long expandSize,long stackSize)
{
    BobInterpreter *c;

    /* allocate the interpreter structure */
    if ((c = (BobInterpreter *)malloc(sizeof(BobInterpreter))) == NULL)
        return NULL;

    /* initialize */
    memset(c,0,sizeof(BobInterpreter));
    c->expandSize = expandSize;

    /* make the initial old memory space */
    if ((c->oldSpace = NewMemorySpace(c,size)) == NULL) {
        free(c);
        return NULL;
    }
    c->pNextOld = &c->oldSpace->next;

    /* make the initial new memory space */
    if ((c->newSpace = NewMemorySpace(c,size)) == NULL) {
        BobFree(c,c->oldSpace);
        c->oldSpace = NULL;
        free(c);
        return NULL;
    }
    c->pNextNew = &c->newSpace->next;

    /* make the stack */
    if ((c->stack = (BobValue *)BobAlloc(c,(size_t)(stackSize * sizeof(BobValue)))) == NULL) {
        BobFree(c,c->oldSpace);
        BobFree(c,c->newSpace);
        free(c);
        return NULL;
    }

    /* initialize the stack */
    c->stackTop = c->stack + stackSize;
    c->sp = c->fp = c->stackTop;

    /* make the symbol table */
    c->symbols = BobMakeObject(c,c->nilValue);

    /* make the nil symbol */
    c->nilValue = BobMakeSymbol(c,"nil",3);

	/* fixup the symbol table */
	BobSetObjectClass(c->symbols,c->nilValue);
    BobSetObjectProperties(c->symbols,c->nilValue);

    /* initialize the global scope */
    c->globalScope.c = c;
    c->globalScope.globals = BobMakeObject(c,c->nilValue);
    c->globalScope.next = c->scopes;
    c->scopes = &c->globalScope;

    /* initialize the current scope */
    c->currentScope.c = c;
    c->currentScope.globals = c->globalScope.globals;
    c->currentScope.next = c->scopes;
    c->scopes = &c->currentScope;

	/* enter nil into the symbol table */
    BobSetGlobalValue(BobCurrentScope(c),c->nilValue,c->nilValue);

    /* make the true and false symbols */
    c->trueValue = BobInternCString(c,"true");
    BobSetGlobalValue(BobCurrentScope(c),c->trueValue,c->trueValue);
    c->falseValue = c->nilValue;
    BobEnterVariable(BobGlobalScope(c),"false",c->falseValue);

    /* create the base of the object heirarchy */
    BobInitObject(c);

    /* initialize the built-in types */
    BobInitTypes(c);
    BobInitMethod(c);
    BobInitVector(c);
    BobInitSymbol(c);
    BobInitString(c);
    BobInitInteger(c);
    BobInitFloat(c);
    BobAddTypeSymbols(c);

    /* initialize the external types */
    BobInitFile(c);

    /* initialize the interpreter */
    InitInterpreter(c);

    /* initialize the standard i/o streams */
    c->standardInput = &BobNullStream;
    c->standardOutput = &BobNullStream;
    c->standardError = &BobNullStream;

    /* return successfully */
    return c;
}

/* BobFreeInterpreter - free an interpreter structure */
void BobFreeInterpreter(BobInterpreter *c)
{
    BobMemorySpace *space,*nextSpace;
    BobProtectedPtrs *p,*nextp;
    BobDispatch *d,*nextd;
    BobScope *s,*nexts;
	BobSymbolBlock *b,*nextb;

    /* destroy cobjects */
    /*BobDestroyAllCObjects(c)*/;

    /* free memory spaces */
    for (space = c->oldSpace; space != NULL; space = nextSpace) {
        nextSpace = space->next;
        BobFree(c,space);
    }
    for (space = c->newSpace; space != NULL; space = nextSpace) {
        nextSpace = space->next;
        BobFree(c,space);
    }

    /* free the stack */
    BobFree(c,c->stack);
    
    /* free the type list */
    for (d = c->types; d != NULL; d = nextd) {
        nextd = d->next;
        BobFreeDispatch(c,d);
    }

    /* free the protected pointer blocks */
    for (p = c->protectedPtrs; p != NULL; p = nextp) {
        nextp = p->next;
        BobFree(c,p);
    }

    /* free the variable scopes */
    for (s = c->scopes; s != NULL; s = nexts) {
        nexts = s->next;
        if (s != &c->globalScope && s != &c->currentScope)
            BobFree(c,s);
    }

    /* free the symbols */
	for (b = c->symbolSpace; b != NULL; b = nextb) {
		nextb = b->next;
		BobFree(c,b);
	}

    /* free the interpreter */
    free(c);
}

/* BobMakeScope - make a variable scope structure */
BobScope *BobMakeScope(BobInterpreter *c,BobScope *parent)
{
    BobScope *scope = (BobScope *)BobAlloc(c,sizeof(BobScope));
    if (scope) {
        scope->c = c;
        scope->globals = BobMakeObject(c,BobScopeObject(parent));
        scope->next = c->scopes;
        c->scopes = scope;
    }
    return scope;
}

/* BobFreeScope - free a variable scope structure */
void BobFreeScope(BobScope *scope)
{
    BobInterpreter *c = scope->c;
    BobScope **pNext,*s;
    if (scope != BobGlobalScope(c) && scope != BobCurrentScope(c)) {
        for (pNext = &c->scopes; (s = *pNext) != NULL; pNext = &s->next)
            if (s == scope) {
                *pNext = s->next;
                break;
            }
        BobFree(c,scope);
    }
}

/* BobInitScope - remove all variables from a scope */
void BobInitScope(BobScope *scope)
{
    BobValue obj = BobScopeObject(scope);
    BobSetObjectProperties(obj,scope->c->nilValue);
    BobSetObjectPropertyCount(obj,0);
}

/* InitInterpreter - initialize an interpreter structure */
static void InitInterpreter(BobInterpreter *c)
{
    /* initialize the stack */
    c->sp = c->fp = c->stackTop;

    /* initialize the registers */
    c->val = c->nilValue;
    c->env = c->nilValue;
    c->code = NULL;
}

/* NewMemorySpace - make a new semi-space */
static BobMemorySpace *NewMemorySpace(BobInterpreter *c,size_t size)
{
    BobMemorySpace *space;
    if ((space = (BobMemorySpace *)BobAlloc(c,(size_t)(sizeof(BobMemorySpace) + size))) != NULL) {
        space->base = (unsigned char *)space + sizeof(BobMemorySpace);
        space->free = space->base;
        space->top = space->base + size;
        space->cObjects = NULL;
        space->next = NULL;
    }
    return space;
}

/* BobAllocate - allocate memory for a value */
BobValue BobAllocate(BobInterpreter *c,long size)
{
    BobMemorySpace *newSpace;
#if 0
    BobMemorySpace *oldSpace;
    long expandSize;
#endif
    BobValue val;

    /* look for free space */
    for (newSpace = c->newSpace; newSpace != NULL; newSpace = newSpace->next)
        if (newSpace->free + size <= newSpace->top) {
            val = (BobValue)newSpace->free;
            newSpace->free += size;
            return val;
        }

    /* collect garbage */
    BobCollectGarbage(c);

    /* look again */
    for (newSpace = c->newSpace; newSpace != NULL; newSpace = newSpace->next)
        if (newSpace->free + size <= newSpace->top) {
            val = (BobValue)newSpace->free;
            newSpace->free += size;
            return val;
        }

#if 0
    /* check for being able to expand memory */
    if ((expandSize = c->expandSize) == 0)
        BobInsufficientMemory(c);
        
    /* make sure we allocate at least as much as needed */
    if (size > expandSize)
        expandSize = size;
    
    /* allocate more old space */
    if (!(oldSpace = NewMemorySpace(c,expandSize)))
        BobInsufficientMemory(c);
    
    /* allocate more new space */
    if (!(newSpace = NewMemorySpace(c,expandSize))) {
        BobFree(c,oldSpace);
        BobInsufficientMemory(c);
    }

    /* link in the new spaces */
    *c->pNextOld = oldSpace;
    c->pNextOld = &oldSpace->next;
    *c->pNextNew = newSpace;
    c->pNextNew = &newSpace->next;

    /* return some of the newly allocated space */
    val = (BobValue)newSpace->free;
    newSpace->free += size;
    return val;
#else
    BobInsufficientMemory(c);
    return c->nilValue; /* never reached */
#endif
}

/* BobMakeDispatch - make a new type dispatch */
BobDispatch *BobMakeDispatch(BobInterpreter *c,char *typeName,BobDispatch *prototype)
{
    int totalSize = sizeof(BobDispatch) + strlen(typeName) + 1;
    BobDispatch *d;

    /* allocate a new dispatch structure */
    if ((d = (BobDispatch *)BobAlloc(c,totalSize)) == NULL) {
        BobDrop(c,1);
        return NULL;
    }

    /* initialize */
    *d = *prototype;
    d->baseType = d;
    d->typeName = (char *)d + sizeof(BobDispatch);
    strcpy(d->typeName,typeName);
    d->object = NULL;

    /* add the type to the type list */
    d->next = c->types;
    c->types = d;

    /* return the new type */
    return d;
}

/* BobFreeDispatch - free a type dispatch structure */
void BobFreeDispatch(BobInterpreter *c,BobDispatch *d)
{
    BobFree(c,d);
}

/* BobCollectGarbage - garbage collect a heap */
void BobCollectGarbage(BobInterpreter *c)
{
    BobProtectedPtrs *ppb;
    BobValue obj,*fp,*p;
    unsigned char *scan;
    BobMemorySpace *ms;
    BobSavedState *ss;
    BobScope *scope;
    BobDispatch *d;

	BobStreamPutS("[GC",c->standardError);

    /* reverse the memory spaces */
    ms = c->oldSpace;
    c->oldSpace = c->newSpace;
    c->newSpace = ms;

    /* reset the new space pointers */
    for (; ms != NULL; ms = ms->next)
        ms->free = ms->base;
    
    /* copy the root objects */
    c->nilValue = BobCopyValue(c,c->nilValue);
    c->trueValue = BobCopyValue(c,c->trueValue);
    c->falseValue = BobCopyValue(c,c->falseValue);
    c->symbols = BobCopyValue(c,c->symbols);
    c->objectValue = BobCopyValue(c,c->objectValue);

    /* copy global variable scopes */
    for (scope = c->scopes; scope != NULL; scope = scope->next)
        scope->globals = BobCopyValue(c,scope->globals);

    /* copy basic types */
    c->methodObject = BobCopyValue(c,c->methodObject);
    c->vectorObject = BobCopyValue(c,c->vectorObject);
    c->symbolObject = BobCopyValue(c,c->symbolObject);
    c->stringObject = BobCopyValue(c,c->stringObject);
    c->integerObject = BobCopyValue(c,c->integerObject);
    c->floatObject = BobCopyValue(c,c->floatObject);

    /* copy the type list */
    for (d = c->types; d != NULL; d = d->next) {
        if (d->object)
            d->object = BobCopyValue(c,d->object);
    }

    /* copy protected pointers */
    for (ppb = c->protectedPtrs; ppb != NULL; ppb = ppb->next) {
        BobValue **pp = ppb->pointers;
        int count = ppb->count;
        for (; --count >= 0; ++pp)
            **pp = BobCopyValue(c,**pp);
    }
    
    /* copy the saved interpreter states */
    for (ss = c->savedState; ss != NULL; ss = ss->next) {
        ss->globals = BobCopyValue(c,ss->globals);
        ss->env = BobCopyValue(c,ss->env);
        if (ss->code != NULL)
            ss->code = BobCopyValue(c,ss->code);
    }
    
    /* copy the stack */
    for (fp = c->fp, p = c->sp; p < c->stackTop; )
        if (p >= fp)
            p = (*((BobCDispatch *)*fp)->copy)(c,&fp);
        else {
            *p = BobCopyValue(c,*p);
            ++p;
        }
        
    /* copy the current code object */
    if (c->code)
        c->code = BobCopyValue(c,c->code);
    
    /* copy the registers */
    c->val = BobCopyValue(c,c->val);
    c->env = BobCopyValue(c,c->env);
        
    /* copy any user objects */
    if (c->protectHandler)
        (*c->protectHandler)(c,c->protectData);

    /* scan and copy until all accessible objects have been copied */
    scan = c->newSpace->base;
	while (scan < c->newSpace->free) {
        obj = (BobValue)scan;
        scan += ValueSize(obj);
        ScanValue(c,obj);
    }
    
    /* fixup cbase and pc */
    if (c->code) {
        long pcoff = c->pc - c->cbase;
        c->cbase = BobStringAddress(BobCompiledCodeBytecodes(c->code));
        c->pc = c->cbase + pcoff;
    }

    {
		char buf[128];
		sprintf(buf,
				" - %ld bytes free, total memory %lu, allocations %lu]\n",
				c->newSpace->top - c->newSpace->free,
				c->totalMemory,
				c->allocCount);
		BobStreamPutS(buf,c->standardError);
	}
      
    /* destroy any unreachable cobjects */
    BobDestroyUnreachableCObjects(c);
}

/* default handlers */

/* BobDefaultGetProperty - get the value of a property */
int BobDefaultGetProperty(BobInterpreter *c,BobValue obj,BobValue tag,BobValue *pValue)
{
    return FALSE;
}

/* BobDefaultSetProperty - set the value of a property */
int BobDefaultSetProperty(BobInterpreter *c,BobValue obj,BobValue tag,BobValue value)
{
	return FALSE;
}

/* BobDefaultNewInstance - create a new instance */
BobValue BobDefaultNewInstance(BobInterpreter *c,BobValue parent)
{
	BobCallErrorHandler(c,BobErrNewInstance,parent);
    return c->nilValue; /* never reached */
}

/* BobDefaultPrint - print an object */
int BobDefaultPrint(BobInterpreter *c,BobValue obj,BobStream *s)
{
    char buf[64];
    sprintf(buf,"%08lx",(long)obj);
    return BobStreamPutC('<',s) == '<'
        && BobStreamPutS(BobTypeName(obj),s) == 0
        && BobStreamPutC('-',s) == '-'
        && BobStreamPutS(buf,s) == 0
        && BobStreamPutC('>',s) == '>';
}

/* BobDefaultCopy - copy an object from old space to new space */
BobValue BobDefaultCopy(BobInterpreter *c,BobValue obj)
{
    long size = ValueSize(obj);
    BobValue newObj;
    
    /* find a place to put the new object */
    newObj = (BobValue)c->newSpace->free;
    
    /* copy the object */
    memcpy(newObj,obj,(size_t)size);
    c->newSpace->free += size;
    
    /* store a forwarding address in the old object */
    BobSetDispatch(obj,&BobBrokenHeartDispatch);
    BobBrokenHeartSetForwardingAddr(obj,newObj);
    
    /* return the new object */return newObj;
}

/* BobBobDefaultScan - scan an object without embedded pointers */
void BobDefaultScan(BobInterpreter *c,BobValue obj)
{
}

/* BobDefaultHash - default object hash routine */
BobIntegerType BobDefaultHash(BobValue obj)
{
    return 0;
}

/* BROKEN HEART */

static long BrokenHeartSize(BobValue obj);
static BobValue BrokenHeartCopy(BobInterpreter *c,BobValue obj);

BobDispatch BobBrokenHeartDispatch = {
    "BrokenHeart",
    &BobBrokenHeartDispatch,
    BobDefaultGetProperty,
    BobDefaultSetProperty,
    BobDefaultNewInstance,
    BobDefaultPrint,
    BrokenHeartSize,
    BrokenHeartCopy,
    BobDefaultScan,
    BobDefaultHash
};

static long BrokenHeartSize(BobValue obj)
{
    return sizeof(BobBrokenHeart);
}

static BobValue BrokenHeartCopy(BobInterpreter *c,BobValue obj)
{
    return BobBrokenHeartForwardingAddr(obj);
}

/* BobProtectPointer - protect a pointer from the garbage collector */
int BobProtectPointer(BobInterpreter *c,BobValue *pp)
{
    BobProtectedPtrs *ppb = c->protectedPtrs;

    /* look for space in an existing block */
    while (ppb) {
        if (ppb->count < BobPPSize) {
            ppb->pointers[ppb->count++] = pp;
            *pp = c->nilValue;
            return TRUE;
        }
        ppb = ppb->next;
    }

    /* allocate a new block */
    if ((ppb = (BobProtectedPtrs *)BobAlloc(c,sizeof(BobProtectedPtrs))) == NULL)
        return FALSE;

    /* initialize the new block */
    ppb->next = c->protectedPtrs;
    c->protectedPtrs = ppb;
    ppb->count = 0;

    /* add the pointer to the new block */
    ppb->pointers[ppb->count++] = pp;
    *pp = c->nilValue;
    return TRUE;
}

/* BobUnprotectPointer - unprotect a pointer from the garbage collector */
int BobUnprotectPointer(BobInterpreter *c,BobValue *pp)
{
    BobProtectedPtrs *ppb = c->protectedPtrs;
    while (ppb) {
        BobValue **ppp = ppb->pointers;
        int cnt = ppb->count;
        while (--cnt >= 0) {
            if (*ppp++ == pp) {
                while (--cnt >= 0) {
                    ppp[-1] = *ppp;
                    ++ppp;
                }
                --ppb->count;
                return TRUE;
            }
        }
        ppb = ppb->next;
    }
    return FALSE;
}

/* BobAlloc - allocate memory */
void *BobAlloc(BobInterpreter *c,unsigned long size)
{
    unsigned long totalSize = sizeof(unsigned long) + size;
    unsigned long *p = malloc(totalSize);
    if (p) {
        *p++ = totalSize;
        c->totalMemory += totalSize;
        ++c->allocCount;
    }
    return (void *)p;
}

/* BobFree - free memory */
void BobFree(BobInterpreter *c,void *p)
{
    unsigned long totalSize = *--(unsigned long *)p;
    if (c) {
        c->totalMemory -= totalSize;
        --c->allocCount;
    }
    free(p);
}

/* BobInsufficientMemory - report an insufficient memory error */
void BobInsufficientMemory(BobInterpreter *c)
{
    BobCallErrorHandler(c,BobErrInsufficientMemory,0);
}
